# Written by Bram Cohen
# multitracker extensions by John Hoffman
# Modified by Cameron Dale
# see LICENSE.txt for license information

# $Id: makemetafile.py 119 2007-06-20 07:40:58Z camrdale-guest $

from os.path import getsize, split, join, abspath, isdir
from os import listdir
from sha import sha
from copy import copy
from string import strip
from DebTorrent.bencode import bencode
from btformats import check_info
from threading import Event
from time import time
from traceback import print_exc
from DebTorrent.zurllib import urlopen
from gzip import GzipFile
from StringIO import StringIO
from re import subn
import binascii
try:
    from sys import getfilesystemencoding
    ENCODING = getfilesystemencoding()
except:
    from sys import getdefaultencoding
    ENCODING = getdefaultencoding()

defaults = [
    ('announce_list', '',
        'a list of announce URLs - explained below'),
    ('deb_mirrors', '',
        'a list of mirror URLs - explained below'),
    ('piece_size_pow2', 0,
        "which power of 2 to set the piece size to (0 = automatic)"),
    ('comment', '',
        "optional human-readable comment to put in .dtorrent"),
    ('filesystem_encoding', '',
        "optional specification for filesystem encoding " +
        "(set automatically in recent Python versions)"),
    ('target', '',
        "optional target file for the torrent"),
    ('pieces_file', '', 'the file that contains the sub-package piece information'),
    ('separate_all', 0, 'create a separate torrent for the architecture:all packages'),
    ]

default_piece_len_exp = 18

ignore = ['core', 'CVS']

def print_announcelist_details():
    print ('    announce_list = optional list of redundant/backup tracker URLs, in the format:')
    print ('           url[,url...][|url[,url...]...]')
    print ('                where URLs separated by commas are all tried first')
    print ('                before the next group of URLs separated by the pipe is checked.')
    print ("                If none is given, it is assumed you don't want one in the metafile.")
    print ('                If announce_list is given, clients which support it')
    print ('                will ignore the <announce> value.')
    print ('           Examples:')
    print ('                http://tracker1.com|http://tracker2.com|http://tracker3.com')
    print ('                     (tries trackers 1-3 in order)')
    print ('                http://tracker1.com,http://tracker2.com,http://tracker3.com')
    print ('                     (tries trackers 1-3 in a randomly selected order)')
    print ('                http://tracker1.com|http://backup1.com,http://backup2.com')
    print ('                     (tries tracker 1 first, then tries between the 2 backups randomly)')
    print ('')
    print ('    deb_mirrors = optional list of mirror URLs, in the format:')
    print ('            url[|url...]')
    
def uniconvertl(l, e):
    r = []
    try:
        for s in l:
            r.append(uniconvert(s, e))
    except UnicodeError:
        raise UnicodeError('bad filename: '+join(l))
    return r

def uniconvert(s, e = None):
    if not e:
        e = ENCODING
    if not e:
        e = 'ascii'
    
    try:
        s = unicode(s,e)
    except UnicodeError:
        raise UnicodeError('bad filename: '+s)
    return s.encode('utf-8')

def convert_all(f):
    # Find the architecture and replace it with 'all'
    (f_all, n) = subn(r'binary-[a-zA-Z0-9]+([^a-zA-Z0-9]?)', r'binary-all\1', f)
    if n == 0:
        # Otherwise add '-all' before the extension
        (f_all, n) = subn(r'\.([^.]*)$', r'-all.\1', f)
        if n == 0:
            # Otherwise add '-all' to the end
            f_all = f + '-all'
    return f_all

def make_meta_file(file, url, params = {}, progress = lambda x: None):
    if params.has_key('piece_size_pow2'):
        piece_len_exp = params['piece_size_pow2']
    else:
        piece_len_exp = default_piece_len_exp
    if params.has_key('target') and params['target'] != '':
        f = params['target']
    else:
        a, b = split(file)
        if b == '':
            f = a + '.dtorrent'
            name = a
        else:
            f = join(a, b + '.dtorrent')
            name = b
            
    if piece_len_exp == 0:  # automatic
        size = calcsize(file)
        if   size > 8L*1024*1024*1024:   # > 8 gig =
            piece_len_exp = 21          #   2 meg pieces
        elif size > 2*1024*1024*1024:   # > 2 gig =
            piece_len_exp = 20          #   1 meg pieces
        elif size > 512*1024*1024:      # > 512M =
            piece_len_exp = 19          #   512K pieces
        elif size > 64*1024*1024:       # > 64M =
            piece_len_exp = 18          #   256K pieces
        elif size > 16*1024*1024:       # > 16M =
            piece_len_exp = 17          #   128K pieces
        elif size > 4*1024*1024:        # > 4M =
            piece_len_exp = 16          #   64K pieces
        else:                           # < 4M =
            piece_len_exp = 15          #   32K pieces
    piece_length = 2 ** piece_len_exp

    encoding = None
    if params.has_key('filesystem_encoding'):
        encoding = params['filesystem_encoding']

    (info, info_all) = makeinfo(file, piece_length, encoding, progress, 
                                params['separate_all'], params['pieces_file'])

    if info:
        create_file(f, info, url, uniconvert(name, encoding), params)
        
    if info_all:
        create_file(convert_all(f), info_all, url, uniconvert(convert_all(name), encoding), params)
        
def create_file(f, info, url, name, params):
    check_info(info)
    h = open(f, 'wb')
    data = {'info': info, 'announce': strip(url), 
        'name': name,
        'creation date': long(time())}
    
    if params.has_key('comment') and params['comment']:
        data['comment'] = params['comment']
        
    if params.has_key('real_announce_list'):    # shortcut for progs calling in from outside
        data['announce-list'] = params['real_announce_list']
    elif params.has_key('announce_list') and params['announce_list']:
        l = []
        for tier in params['announce_list'].split('|'):
            l.append(tier.split(','))
        data['announce-list'] = l
        
    if params.has_key('real_deb_mirrors'):    # shortcut for progs calling in from outside
        data['deb_mirrors'] = params['real_deb_mirrors']
    elif params.has_key('deb_mirrors') and params['deb_mirrors']:
        data['deb_mirrors'] = params['deb_mirrors'].split('|')
        
    h.write(bencode(data))
    h.close()

def calcsize(file):
    if not isdir(file):
        return getsize(file)
    total = 0L
    for s in subfiles(abspath(file)):
        total += getsize(s[1])
    return total

def getsubpieces(file, pieces_file = ''):
    pieces = {}
    packages = 0

    if pieces_file:
        try:
            f = open(pieces_file)
        except:
            return {}
    elif 'dists' in file.split('_'):
        try:
            parts = file.split('_')
            try:
                parts[parts.index('stable', parts.index('dists'))] = 'etch'
            except:
                pass
            try:
                parts[parts.index('testing', parts.index('dists'))] = 'lenny'
            except:
                pass
            try:
                parts[parts.index('unstable', parts.index('dists'))] = 'sid'
            except:
                pass
            piece_url = 'http://merkel.debian.org/~ajt/extrapieces/dists_'
            piece_url += '_'.join(parts[parts.index('dists')+1:])
            if piece_url.endswith('.gz'):
                piece_url = piece_url[:-3]
            if piece_url.endswith('.bz2'):
                piece_url = piece_url[:-4]
            piece_url += '-extrapieces.gz'
            piece_file = urlopen(piece_url)
            piece_data = piece_file.read()
            try:
                piece_file.close()
            except:
                pass
            f = piece_data.split('\n')
        except:
            return {}
    else:
        return {}

    p = [None, [], []]
    read_data = False
    for line in f:
        line = line.rstrip()
        if line == "":
            if (p[0] and p[1] and p[2]):
                pieces[p[0]] = (p[1], p[2])
                packages += 1
                #progress(packages)
            p = [None, [], []]
            read_data = False
        if read_data == True and line[:1] != " ":
            read_data = False
        if line[:9] == "Filename:":
            p[0] = line[10:]
        if line == "SHA1-Pieces:":
            read_data = True
        if read_data == True and line[:1] == " ":
            p[1].append(binascii.a2b_hex(line[1:41]))
            p[2].append(int(line[42:]))

    try:
        f.close()
    except:
        pass

    return pieces

def getpieces(f, encoding = None, progress = lambda x: None, separate_all = 0, sub_pieces = {}):
    
    if not encoding:
        encoding = ENCODING
    if not encoding:
        encoding = 'ascii'
    
    pieces = ([], [])
    lengths = ([], [])
    fs = ([], [])
    packages = [0, 0]
    info = None
    info_all = None
    
    p = [None, None, None, None, None, None]
    for line in f:
        line = line.rstrip()
        if line == "":
            if (p[0] and p[1] and p[2]):
                # Check which torrent to add the info to
                all = 0
                if (separate_all and p[5] == 'all'):
                    all = 1

                if sub_pieces.has_key(p[1]):
                    lengths[all].extend(sub_pieces[p[1]][1])
                    pieces[all].extend(sub_pieces[p[1]][0])
                else:
                    lengths[all].append(p[0])
                    pieces[all].append(p[2])

                path = []
                while p[1]:
                    p[1],d = split(p[1])
                    path.insert(0,d)
                fs[all].append({'length': p[0], 'path': uniconvertl(path, encoding)})
                packages[all] += 1
                progress(packages[0] + packages[1])
            p = [None, None, None, None, None, None]
        if line[:9] == "Filename:":
            p[1] = line[10:]
        if line[:5] == "Size:":
            p[0] = long(line[6:])
        if line[:5] == "SHA1:":
            p[2] = binascii.a2b_hex(line[6:])
        if line[:8] == "Package:":
            p[3] = line[9:]
        if line[:8] == "Version:":
            p[4] = line[9:]
        if line[:13] == "Architecture:":
            p[5] = line[14:]

    if packages[0] > 0:
        info = {'pieces': ''.join(pieces[0]), 'piece lengths': lengths[0], 'files': fs[0]}
    if packages[1] > 0:
        info_all = {'pieces': ''.join(pieces[1]), 'piece lengths': lengths[1], 'files': fs[1]}

    return (info, info_all)

def makeinfo(file, piece_length, encoding, progress, separate_all = 0, pieces_file = ''):

    sub_pieces = getsubpieces(file, pieces_file)

    if not sub_pieces:
        print 'WARNING: Pieces file could not be found, not using sub-package pieces.'
        
    file = abspath(file)
    f = open(file)
    (info, info_all) = getpieces(f, encoding, progress, separate_all = separate_all, sub_pieces = sub_pieces)
    f.close()
    
    return (info, info_all)

def subfiles(d):
    r = []
    stack = [([], d)]
    while len(stack) > 0:
        p, n = stack.pop()
        if isdir(n):
            for s in listdir(n):
                if s not in ignore and s[:1] != '.':
                    stack.append((copy(p) + [s], join(n, s)))
        else:
            r.append((p, n))
    return r


def completedir(dir, url, params = {}, vc = lambda x: None, fc = lambda x: None):
    files = listdir(dir)
    files.sort()
    ext = '.dtorrent'
    if params.has_key('target'):
        target = params['target']
    else:
        target = ''

    togen = []
    for f in files:
        if f[-len(ext):] != ext and (f + ext) not in files:
            togen.append(join(dir, f))
        
    for i in togen:
        fc(i)
        try:
            t = split(i)[-1]
            if t not in ignore and t[0] != '.':
                if target != '':
                    params['target'] = join(target,t+ext)
                make_meta_file(i, url, params, progress = vc)
        except ValueError:
            print_exc()
