#!/usr/bin/env python
#
# Written by John Hoffman
# Modified by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: launchmanycore.py 129 2007-06-27 04:05:43Z camrdale-guest $

"""Manage the downloading of multiple torrents in one process."""

from DebTorrent import PSYCO
if PSYCO.psyco:
    try:
        import psyco
        assert psyco.__version__ >= 0x010100f0
        psyco.full()
    except:
        pass

from download_bt1 import BT1Download
from RawServer import RawServer, UPnP_ERROR
from RateLimiter import RateLimiter
from ServerPortHandler import MultiHandler
from natpunch import UPnP_test
from random import seed
from socket import error as socketerror
from threading import Event
from sys import argv, exit
import sys, os, binascii
from clock import clock
from __init__ import createPeerID, mapbase64, version
from cStringIO import StringIO
from traceback import print_exc
from DebTorrent.BT1.AptListener import AptListener
from DebTorrent.HTTPHandler import HTTPHandler

try:
    True
except:
    True = 1
    False = 0

DEBUG = True

def fmttime(n):
    """Formats seconds into a human-readable time.
    
    Formats a given number of seconds into a human-readable time appropriate
    for display to the user.
    
    @type n: C{int}
    @param n: the number of seconds
    @rtype: C{string}
    @return: a displayable representation of the number of seconds
    
    """
    
    try:
        n = int(n)  # n may be None or too large
        assert n < 5184000  # 60 days
    except:
        return 'downloading'
    m, s = divmod(n, 60)
    h, m = divmod(m, 60)
    return '%d:%02d:%02d' % (h, m, s)

class SingleDownload:
    """Manage a single torrent download.
    
    @type controller: L{LaunchMany}
    @ivar controller: the manager for all torrent downloads
    @type hash: C{string}
    @ivar hash: the info hash of the torrent
    @type response: C{dictionary}
    @ivar response: the meta info for the torrent
    @type config: C{dictionary}
    @ivar config: the configuration parameters
    @type doneflag: C{threading.Event}
    @ivar doneflag: the flag that indicates when the torrent is to be shutdown
    @type waiting: C{boolean}
    @ivar waiting: unknown
    @type checking: C{boolean}
    @ivar checking: unknown
    @type working: C{boolean}
    @ivar working: unknown
    @type seed: C{boolean}
    @ivar seed: unknown
    @type closed: C{boolean}
    @ivar closed: unknown
    @type status_msg: C{string}
    @ivar status_msg: the current activity the torrent is engaged in
    @type status_err: C{list} of C{string}
    @ivar status_err: the list of errors that have occurred
    @type status_errtime: C{int}
    @ivar status_errtime: the time of the last error
    @type status_done: C{float}
    @ivar status_done: the fraction of the current activity that is complete
    @type rawserver: L{ServerPortHandler.SingleRawServer}
    @ivar rawserver: the simplified Server to use to handle this torrent
    @type d: L{download_bt1.BT1Download}
    @ivar d: the downloader for the torrent
    @type _hashcheckfunc: C{method}
    @ivar _hashcheckfunc: the method to call to hash check the torrent
    @type statsfunc: C{method}
    @ivar statsfunc: the method to call to get the statistics for the running download
    
    """
    
    def __init__(self, controller, hash, response, config, myid):
        """Initialize the instance and start a new downloader.
        
        @type controller: L{LaunchMany}
        @param controller: the manager for all torrent downloads
        @type hash: C{string}
        @param hash: the info hash of the torrent
        @type response: C{dictionary}
        @param response: the meta info for the torrent
        @type config: C{dictionary}
        @param config: the configuration parameters
        @type myid: C{string}
        @param myid: the peer ID to use
        
        """
        
        self.controller = controller
        self.hash = hash
        self.response = response
        self.config = config
        
        self.doneflag = Event()
        self.waiting = True
        self.checking = False
        self.working = False
        self.seed = False
        self.closed = False

        self.status_msg = ''
        self.status_err = ['']
        self.status_errtime = 0
        self.status_done = 0.0

        self.rawserver = controller.handler.newRawServer(hash, self.doneflag)

        d = BT1Download(self.display, self.finished, self.error,
                        controller.exchandler, self.doneflag, config, response,
                        hash, myid, self.rawserver, controller.listen_port, 
                        self.controller.configdir)
        self.d = d

    def start(self):
        """Initialize the new torrent download and schedule it for hash checking."""
        if not self.d.saveAs(self.saveAs):
            self._shutdown()
            return
        self._hashcheckfunc = self.d.initFiles()
        if not self._hashcheckfunc:
            self._shutdown()
            return
        self.controller.hashchecksched(self.hash)


    def saveAs(self, name, length, saveas, isdir):
        """Determine the location to save the torrent in.
        
        @type name: C{string}
        @param name: the name from the torrent's metainfo
        @type length: C{long}
        @param length: the total length of the torrent download (not used)
        @type saveas: C{string}
        @param saveas: the user specified location to save to
        @type isdir: C{boolean}
        @param isdir: whether the torrent needs a directory
        @rtype: C{string}
        @return: the location to save the torrent in
        
        """
        
        return self.controller.saveAs(self.hash, name, saveas, isdir)

    def hashcheck_start(self, donefunc):
        """Start the hash checking of the torrent.
        
        @type donefunc: C{method}
        @param donefunc: the method to call when the hash checking is complete
        
        """
        
        if self.is_dead():
            self._shutdown()
            return
        self.waiting = False
        self.checking = True
        self._hashcheckfunc(donefunc)

    def hashcheck_callback(self):
        """Start the torrent running now that hash checking is complete."""
        self.checking = False
        if self.is_dead():
            self._shutdown()
            return
        if not self.d.startEngine(ratelimiter = self.controller.ratelimiter):
            self._shutdown()
            return
        self.d.startRerequester()
        self.statsfunc = self.d.startStats()
        self.rawserver.start_listening(self.d.getPortHandler())
        self.working = True

    def is_dead(self):
        """Check if the torrent download has been shutdown.
        
        @rtype: C{boolean}
        @return: whether the torrent download has been shutdown
        
        """
        
        return self.doneflag.isSet()

    def _shutdown(self):
        """Loudly shutdown the running torrent."""
        self.shutdown(False)

    def shutdown(self, quiet=True):
        """Shutdown the running torrent.
        
        @type quiet: C{boolean}
        @param quiet: whether to announce the shutdown (optional, defaults to True)
        
        """
        
        if self.closed:
            return
        self.doneflag.set()
        self.rawserver.shutdown()
        if self.checking or self.working:
            self.d.shutdown()
        self.waiting = False
        self.checking = False
        self.working = False
        self.closed = True
        self.controller.was_stopped(self.hash)
        if not quiet:
            self.controller.died(self.hash)
            

    def display(self, activity = None, fractionDone = None):
        """Update the current activity's status for later display.
        
        Really only used by StorageWrapper now.
        
        @type activity: C{string}
        @param activity: the activity currently under way 
            (optional, defaults to not changing the current activity)
        @type fractionDone: C{float}
        @param fractionDone: the fraction of the activity that is complete
            (optional, defaults to not changing the current fraction done)
        
        """
        
        if activity:
            self.status_msg = activity
        if fractionDone is not None:
            self.status_done = float(fractionDone)

    def finished(self):
        """Indicate that the download has completed."""
        self.seed = True

    def error(self, msg):
        """Add a new error to the list of errors that have occurred.
        
        @type msg: C{string}
        @param msg: the error message
        
        """
        
        if self.doneflag.isSet():
            self._shutdown()
        self.status_err.append(msg)
        self.status_errtime = clock()


class LaunchMany:
    """Manage the collection of all single torrent downloads.
    
    @type config: C{dictionary}
    @ivar config: the configuration parameters
    @type Output: unknown
    @ivar Output: the displayer instance to use
    @type configdir: L{ConfigDir.ConfigDir}
    @ivar configdir: the configuration and cache directory manager
    @type torrent_cache: C{dictionary}
    @ivar torrent_cache: the cache of known torrents, keys are info hashes
    @type file_cache: C{dictionary}
    @ivar file_cache: the files found in the parsing of the torrent directory
    @type blocked_files: C{dictionary}
    @ivar blocked_files: the torrents in the torrent directory that will not be run
    @type stats_period: C{int}
    @ivar stats_period: the number of seconds between printing the stats for the user
    @type torrent_list: C{list} of C{string}
    @ivar torrent_list: the list of running torrents' info hashes
    @type downloads: C{dictionary}
    @ivar downloads: the currently running downloaders, keys are info hashes
    @type counter: C{int}
    @ivar counter: the number of torrents that have been started so far
    @type doneflag: C{threading.Event}
    @ivar doneflag: flag to indicate all is to be stopped
    @type hashcheck_queue: C{list} of C{string}
    @ivar hashcheck_queue: the list of torrent info hashes waiting to be hash checked
    @type hashcheck_current: C{string}
    @ivar hashcheck_current: the info hash of the torrent currently being hash checked
    @type rawserver: L{RawServer.RawServer}
    @ivar rawserver: the Server instance to use for the downloads
    @type listen_port: C{int}
    @ivar listen_port: the port to listen on for incoming torrent connections
    @type aptlistener: L{BT1.AptListener.AptListener}
    @ivar aptlistener: the AptListener instance used to listen for incoming connections from Apt
    @type ratelimiter: L{RateLimiter.RateLimiter}
    @ivar ratelimiter: the limiter used to cap the maximum upload rate
    @type handler: L{ServerPortHandler.MultiHandler}
    @ivar handler: the multi torrent port listener used to handle connections
    
    """
    
    def __init__(self, config, Output, configdir):
        """Initialize the instance.
        
        @type config: C{dictionary}
        @param config: the configuration parameters
        @type Output: unknown
        @param Output: the displayer instance to use
        @type configdir: L{ConfigDir.ConfigDir}
        @param configdir: the configuration and cache directory manager
        
        """
        
        try:
            self.config = config
            self.Output = Output
            self.configdir = configdir

            self.torrent_cache = {}
            self.file_cache = {}
            self.blocked_files = {}
            self.stats_period = config['display_interval']

            self.torrent_list = []
            self.downloads = {}
            self.counter = 0
            self.doneflag = Event()

            self.hashcheck_queue = []
            self.hashcheck_current = None
            
            self.rawserver = RawServer(self.doneflag, config['timeout_check_interval'],
                              config['timeout'], ipv6_enable = config['ipv6_enabled'],
                              failfunc = self.failed, errorfunc = self.exchandler)
            upnp_type = UPnP_test(config['upnp_nat_access'])
            while True:
                try:
                    self.listen_port = self.rawserver.find_and_bind(
                                    config['minport'], config['maxport'], config['bind'],
                                    ipv6_socket_style = config['ipv6_binds_v4'],
                                    upnp = upnp_type, randomizer = config['random_port'])
                    break
                except socketerror, e:
                    if upnp_type and e == UPnP_ERROR:
                        self.Output.message('WARNING: COULD NOT FORWARD VIA UPnP')
                        upnp_type = 0
                        continue
                    self.failed("Couldn't listen - " + str(e))
                    return

            self.aptlistener = AptListener(self, config, self.rawserver)
            self.rawserver.bind(config['port'], config['bind'],
                   reuse = True, ipv6_socket_style = config['ipv6_binds_v4'])
            self.rawserver.set_handler(HTTPHandler(self.aptlistener.get, 
                                                   config['min_time_between_log_flushes']), 
                                       config['port'])
    
            self.ratelimiter = RateLimiter(self.rawserver.add_task,
                                           config['upload_unit_size'])
            self.ratelimiter.set_upload_rate(config['max_upload_rate'])

            self.handler = MultiHandler(self.rawserver, self.doneflag, config)
            seed(createPeerID())
            self.rawserver.add_task(self.stats, 0)

            # Restore the previous state of the downloads
            self.unpickle(self.configdir.getState())
            
            self.handler.listen_forever()

            # Save the current state of the downloads
            self.configdir.saveState(self.pickle())

            self.Output.message('shutting down')
            self.hashcheck_queue = []
            for hash in self.torrent_list:
                self.Output.message('dropped "'+self.torrent_cache[hash]['path']+'"')
                self.downloads[hash].shutdown()
            self.rawserver.shutdown()

        except:
            data = StringIO()
            print_exc(file = data)
            Output.exception(data.getvalue())


    def stats(self):
        """Call the Output display with the currently running torrents' statistics."""
        self.rawserver.add_task(self.stats, self.stats_period)
        data = self.gather_stats()
        stop = self.Output.display(data)
        if stop:
            self.doneflag.set()

    def gather_stats(self):
        """Gather the statistics for the currently running torrents.
        
        Returns a list, one per running torrent, of tuples:
        
        (C{string}, C{string}, C{string}, C{int}, C{int}, C{string}, C{float},
        C{float}, C{float}, C{long}, C{long}, C{long}, C{float}, C{string})
            
        Which are the name, current status, progress report, number of peers, number of seeds, unknown, number of distributed copies,
        upload rate, download rate, amount uploaded, amount downloaded, total length, time remaining, and latest error message.
        
        @rtype: C{list}
        @return: various statistics for the running torrents
        
        """
        
        data = []
        for hash in self.torrent_list:
            cache = self.torrent_cache[hash]
            if self.config['display_path']:
                name = cache['path']
            else:
                name = cache['name']
            size = cache['length']
            d = self.downloads[hash]
            progress = '0.0%'
            peers = 0
            seeds = 0
            seedsmsg = "S"
            dist = 0.0
            uprate = 0.0
            dnrate = 0.0
            upamt = 0
            dnamt = 0
            t = 0
            if d.is_dead():
                status = 'stopped'
            elif d.waiting:
                status = 'waiting for hash check'
            elif d.checking:
                status = d.status_msg
                progress = '%.1f%%' % (d.status_done*100)
            else:
                stats = d.statsfunc()
                s = stats['stats']
                if d.seed:
                    status = 'seeding'
                    progress = '100.0%'
                    seeds = s.numOldSeeds
                    seedsmsg = "s"
                    dist = s.numCopies
                else:
                    if s.numSeeds + s.numPeers:
                        t = stats['time']
                        if t == 0:  # unlikely
                            t = 0.01
                        status = fmttime(t)
                    else:
                        t = -1
                        status = 'connecting to peers'
                    progress = '%.1f%%' % (int(stats['frac']*1000)/10.0)
                    seeds = s.numSeeds
                    dist = s.numCopies2
                    dnrate = stats['down']
                peers = s.numPeers
                uprate = stats['up']
                upamt = s.upTotal
                dnamt = s.downTotal
                   
            if d.is_dead() or d.status_errtime+300 > clock():
                msg = d.status_err[-1]
            else:
                msg = ''

            data.append(( name, status, progress, peers, seeds, seedsmsg, dist,
                          uprate, dnrate, upamt, dnamt, size, t, msg ))

        return data

    def remove(self, hash):
        """Stop and remove a running torrent.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        
        """
        
        self.torrent_list.remove(hash)
        self.downloads[hash].shutdown()
        del self.downloads[hash]
        
    def add(self, hash, data, save_cache = True):
        """Start a new torrent running, or add the data to the current one.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        @type data: C{dictionary}
        @param data: various info about the torrent, including the metainfo
        @type save_cache: C{boolean}
        @param save_cache: whether to save the torrent metainfo to the cache
        
        """
        
        # If the torrent is already cached, just add new deb_mirrors to it
        new_debmirrors = []
        if hash in self.torrent_cache:
            for u in data['metainfo'].get('deb_mirrors', []):
                if u not in self.torrent_cache[hash]['metainfo'].get('deb_mirrors', []):
                    new_debmirrors.append(u)
                    self.torrent_cache[hash]['metainfo'].setdefault('deb_mirrors', []).append(u)
        else:
            self.torrent_cache[hash] = data

        # Save the new torrent file to the cache
        if save_cache:
            self.configdir.writeTorrent(self.torrent_cache[hash]['metainfo'], hash)

        if hash in self.torrent_list:
            if not self.config['disable_http_downloader']:
                for u in new_debmirrors:
                    self.downloads[hash].d.httpdownloader.make_download(u)
        else:
            c = self.counter
            self.counter += 1
            x = ''
            for i in xrange(3):
                x = mapbase64[c & 0x3F]+x
                c >>= 6
            peer_id = createPeerID(x)
            d = SingleDownload(self, hash, data['metainfo'], self.config, peer_id)
            self.torrent_list.append(hash)
            self.downloads[hash] = d
            d.start()


    def saveAs(self, hash, name, saveas, isdir):
        """Determine the location to save the torrent in.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        @type name: C{string}
        @param name: the name from the torrent's metainfo
        @type saveas: C{string}
        @param saveas: the user specified location to save to
        @type isdir: C{boolean}
        @param isdir: whether the torrent needs a directory
        @rtype: C{string}
        @return: the location to save the torrent in
        
        """
        
        x = self.torrent_cache[hash]
        style = self.config['saveas_style']
        if style == 1:
            if saveas:
                saveas = os.path.join(saveas, name)
            else:
                saveas = name
        else:
            if saveas:
                saveas = os.path.join(saveas, x['mirror'])
            else:
                saveas = x['mirror']
                
        if isdir:
            if not os.path.exists(saveas):
                try:
                    os.makedirs(saveas)
                except:
                    print_exc()
                    raise OSError("couldn't create directory for "+x['path']
                                          +" ("+saveas+")")
        else:
            if not os.path.exists(os.path.split(saveas)[0]):
                try:
                    os.makedirs(os.path.split(saveas)[0])
                except:
                    raise OSError("couldn't create directory for "+x['path']
                                          +" ("+os.path.split(saveas)[0]+")")
            
        return saveas


    def find_file(self, mirror, path):
        """Find which running torrent has the file.
        
        Checks the metainfo of each torrent in the cache to find one that
        has a file whose 'path' matches the given file's path.
        
        @type mirror: C{string}
        @param mirror: mirror name to find the download in
        @type path: C{list} of C{string}
        @param path: the path of the file to find
        @rtype: L{download_bt1.BT1Download}, C{int}
        @return: the running torrent that contains the file and the file's number
            (or None if no running torrents contain the file)
        
        """

        file = '/'.join(path)
        if DEBUG:
            print 'Trying to find file:', file
            
        # Check each torrent in the cache
        for hash, data in self.torrent_cache.items():
            # Make sure this torrent is from the mirror in question 
            if not data['metainfo']['name'].startswith(mirror):
                continue

            file_num = -1
            for f in data['metainfo']['info']['files']:
                file_num += 1
                
                # Check that the file ends with the desired file name
                if file.endswith('/'.join(f['path'])):
                    if DEBUG:
                        print 'Found file in:', binascii.b2a_hex(hash)
                    return self.downloads[hash].d, file_num
                
        if DEBUG:
            print 'Failed to find file.'
        return None, None


    def hashchecksched(self, hash = None):
        """Schedule a new torrent for hash checking.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent to schedule 
            (optional, default is to start the next torrent in the queue)
        
        """
        
        if hash:
            self.hashcheck_queue.append(hash)
        if not self.hashcheck_current:
            self._hashcheck_start()

    def _hashcheck_start(self):
        """Start hash checking the next torrent in the queue."""
        self.hashcheck_current = self.hashcheck_queue.pop(0)
        self.downloads[self.hashcheck_current].hashcheck_start(self.hashcheck_callback)

    def hashcheck_callback(self):
        """Start another torrent's hash check now that the current one is complete."""
        self.downloads[self.hashcheck_current].hashcheck_callback()
        if self.hashcheck_queue:
            self._hashcheck_start()
        else:
            self.hashcheck_current = None

    def died(self, hash):
        """Inform the Output that the torrent has died.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        
        """
        
        if self.torrent_cache.has_key(hash):
            self.Output.message('DIED: "'+self.torrent_cache[hash]['path']+'"')
        
    def has_torrent(self, hash):
        """Determine whether there is a downloader for the torrent.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        @rtype: C{boolean}
        @return: whether the torrent is in the cache of known torrents
        
        """
        
        return self.torrent_cache.has_key(hash)
        
    def was_stopped(self, hash):
        """Remove the torrent from the hash check queue, even if it's already happening.
        
        @type hash: C{string}
        @param hash: the info hash of the torrent
        
        """
        
        try:
            self.hashcheck_queue.remove(hash)
        except:
            pass
        if self.hashcheck_current == hash:
            self.hashcheck_current = None
            if self.hashcheck_queue:
                self._hashcheck_start()

    def failed(self, s):
        """Indicate to the Output that a failure has occurred.
        
        @type s: C{string}
        @param s: the failure message
        
        """
        
        self.Output.message('FAILURE: '+s)

    def exchandler(self, s):
        """Indicate to the Output that an exception has occurred.
        
        @type s: C{string}
        @param s: the exception that occurred
        
        """
        
        self.Output.exception(s)

    def pickle(self):
        """Save the current state of the downloads to a writable state.
        
        Pickled data format::
    
            d['torrent cache'] = {info hash: C{dictionary}, ...}
                        Contains the cached data for all running torrents, keys are 
                        the torrent's info hash, and the data is all the data
                        saved in L{torrent_cache}. There is also a new key added
                        "paused" for torrent's that were running but paused.

        @rtype: C{dictionary}
        @return: the saved state of the current downloads
        
        """
        
        d = {}
        for hash in self.torrent_list:
            d[hash] = {}
            for k in self.torrent_cache[hash].keys():
                if k != 'metainfo':
                    d[hash][k] = self.torrent_cache[hash][k]
            if not self.downloads[hash].d.unpauseflag.isSet():
                d[hash]['paused'] = 1
        return {'torrent cache': d}

    def unpickle(self, data):
        """Restore the current state of the downloads.
        
        Reads the list of previosuly running torrents, loads their metainfo
        from the cache and starts the downloads running. Also pauses the
        download if it was previously paused.
        
        @type data: C{dictionary}
        @param data: the saved state of the previously running downloads downloads
        
        """
        
        if data is None:
            return
        
        d = data['torrent cache']
        for hash in d:
            paused = d[hash].pop('paused', False)
            metainfo = self.configdir.getTorrent(hash)
            if metainfo:
                d[hash]['metainfo'] = metainfo
                self.add(hash, d[hash], False)
                if paused:
                    self.downloads[hash].d.Pause()
