#!/usr/bin/env python
#
# Written by John Hoffman
# Modified by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: btlaunchmany.py 129 2007-06-27 04:05:43Z camrdale-guest $

"""A multiple downloader for debtorrent.

This script implements a command-line client for managing multiple 
debtorrent downloads.

"""

from DebTorrent import PSYCO
if PSYCO.psyco:
    try:
        import psyco
        assert psyco.__version__ >= 0x010100f0
        psyco.full()
    except:
        pass

from DebTorrent.launchmanycore import LaunchMany
from DebTorrent.download_bt1 import defaults, get_usage
from DebTorrent.parseargs import parseargs
from threading import Event
from sys import argv, exit
import sys, os
from DebTorrent import version, report_email
from DebTorrent.ConfigDir import ConfigDir

assert sys.version >= '2', "Install Python 2.0 or greater"
try:
    True
except:
    True = 1
    False = 0

PROFILER = False

def hours(n):
    """Formats seconds into a human-readable time.
    
    Formats a given number of seconds into a human-readable time appropriate
    for display to the user.
    
    @type n: C{int}
    @param n: the number of seconds
    @rtype: C{string}
    @return: a displayable representation of the number of seconds
    
    """
    
    if n == 0:
        return 'complete!'
    try:
        n = int(n)
        assert n >= 0 and n < 5184000  # 60 days
    except:
        return '<unknown>'
    m, s = divmod(n, 60)
    h, m = divmod(m, 60)
    if h > 0:
        return '%d hour %02d min %02d sec' % (h, m, s)
    else:
        return '%d min %02d sec' % (m, s)


Exceptions = []
"""The exceptions that have occurred."""

class HeadlessDisplayer:
    """Displays on the command-line the status of the downloads.
    
    Implements a displayer for showing in text format the current status
    of the downloads.

    """
    
    def display(self, data):
        """Displays the current status of the downloads.
        
        Writes a status update to standard output.
        
        @type data: C{list} of C{tuple}
        @param data: a tuple per download, containing the current status of the download
        @rtype: C{boolean}
        @return: False
        
        """

        print ''
        if not data:
            self.message('no torrents')
        for x in data:
            ( name, status, progress, peers, seeds, seedsmsg, dist,
              uprate, dnrate, upamt, dnamt, size, t, msg ) = x
            print '"%s": "%s" (%s) - %sP%s%s%.3fD u%0.1fK/s-d%0.1fK/s u%dK-d%dK "%s"' % (
                        name, status, progress, peers, seeds, seedsmsg, dist,
                        uprate/1000, dnrate/1000, upamt/1024, dnamt/1024, msg)
        return False
            
    def message(self, s):
        """Prints a message to standard output.
        
        @type s: C{string}
        @param s: the message to print to standard output
        
        """
        
        print "### "+s

    def exception(self, s):
        """Saves a new exception and alerts the user.
        
        @type s: C{string}
        @param s: the exception that occurred
        
        """
        
        Exceptions.append(s)
        self.message('SYSTEM ERROR - EXCEPTION GENERATED')


def run(params):
    """Runs the downloader.
    
    The main function used to create the displayer and start the download.
    
    @type params: C{list} of C{strings}
    @param params: a list of the command-line arguments given to the script
    
    """
    
    defaults.extend( [
        ( 'saveas_style', 1,
          "How to name torrent download directory " + 
          "(1 = <mirror>_dists_<suite>_<section>_binary-<arch>, " +
          "2 = <mirror> (caution: experimental))" ),
        ( 'display_path', 0,
          "whether to display the full path or the torrent contents for each torrent" ),
        ('separate_all',0, 'whether to separate the architecture:all packages, ' +
          "0 = don't separate, 1 = separate and run architecture:all, " +
          "2 = separate and run all architectures but all, " + "3 = separate and run both"),
    ] )
    try:
        try:
            configdir = ConfigDir('launchmany', params[params.index('--config_dir')+1])
        except:
            configdir = ConfigDir('launchmany')
        defaultsToIgnore = ['responsefile', 'url', 'priority']
        configdir.setDefaults(defaults,defaultsToIgnore)
        configdefaults = configdir.loadConfig()
        defaults.append(('save_options',0,
         "whether to save the current options as the new default configuration " +
         "(only for btlaunchmany.py)"))
        config, args = parseargs(params, defaults, 0, 0, configdefaults)
        if config['save_options']:
            configdir.saveConfig(config)
        configdir.deleteOldCacheData(config['expire_cache_data'])
    except ValueError, e:
        print 'error: ' + str(e) + '\n'
        print "Usage: btlaunchmany.py <global options>\n"
        print get_usage(defaults, 80, configdefaults)
        exit(1)

    LaunchMany(config, HeadlessDisplayer(), configdir)
    if Exceptions:
        print '\nEXCEPTION:'
        print Exceptions[0]
        print 'please report this to '+report_email

if __name__ == '__main__':
    if argv[1:] == ['--version']:
        print version
        exit(0)

    if PROFILER:
        import profile, pstats
        p = profile.Profile()
        p.runcall(run, argv[1:])
        p.dump_stats('btlaunchmany.pstat')
#        pstats.Stats(p).strip_dirs().sort_stats('cumulative').print_stats()
        pstats.Stats(p).strip_dirs().sort_stats('time').print_stats()
    else:
        run(argv[1:])
